# This file is part of pymadcad,  distributed under license LGPL v3

''' Definition of 3D primitive objects

	Primitives are parametrized objects, that can be baked into a mesh/web/wire object. A primitive object must have the following signature:
	
		class SomePrimitive:
			# method baking the primitive in some general-purpose 3D object
			def mesh(self) -> Mesh/Web/Wire:
				...
			
			# for the solver
			# primitive attributes the solver has to consider as variables or variable container
			slvvars = 'fields', 'for', 'solver', 'variables'
			# optional method constraining the primitive parameters (to keep points on a circle for instance)
			def fit(self) -> err**2 as float:
				...


	Curve resolution
	----------------

	Some primitive types are curves, the discretisation is important for visual as well as for result quality (remember that even if something looks like a perfect curves, it's still polygons).
	The resolution (subdivision) of curve is done following the following cirterions present in the 'settings' module

	specification priority order:
		
		1. optional argument `resolution` passed to `primitive.mesh()` or to `web()` or `wire()`
		2. optional attribute `resolution` of the primitive object
		3. value of `settings.primitives['curve_resolution']` at bake time.

	specification format:
		
		('fixed', 16)   # fixed amount of 16 subdivisions
		('rad', 0.6)    # max polygon angle is 0.6 rad
		('radm', 0.6)
		('radm2', 0.6)

'''

from math import sqrt
from .mathutils import *
from . import settings
from . import displays
from . import mesh



def isprimitive(obj):
	''' return True if obj match the signature for primitives '''
	return hasattr(obj, 'mesh') and hasattr(obj, 'slvvars')
		
Vector = Point = vec3


class Axis(object):
	''' Mimic the behavior of a tuple, but with the primitive signature. '''
	__slots__ = ('origin', 'direction', 'interval')
	def __init__(self, origin, direction, interval=None):
		self.origin, self.direction = origin, direction
		self.interval = interval
	
	def __getitem__(self, i):
		''' behave like the axis was a tuple (origin, direction) '''
		if i==0:	return self.origin
		elif i==1:	return self.direction
		else:		raise IndexError('an axis has only 2 components')
		
	def flip(self) -> 'Axis':
		''' switch the axis direction '''
		return Axis(self.origin, -self.direction, self.interval)
	
	def offset(self, increment) -> 'Axis':
		''' move the axis origin along its direction '''
		return Axis(self.origin + self.direction*increment, self.direction, self.interval)
		
	def transform(self, transform) -> 'Axis':
		''' move the axis by the given transformation '''
		if isinstance(transform, (float,int)):		return Axis(transform*self.origin, self.direction, self.interval)
		elif isinstance(transform, vec3):			return Axis(transform+self.origin, self.direction, self.interval)
		elif isinstance(transform, (mat3, quat)):	return Axis(transform*self.origin, normalize(transform*self.direction), self.interval)
		elif isinstance(transform, (mat4)):			return Axis(transform*self.origin, normalize(mat3(transform)*self.direction), self.interval)
		raise TypeError('transform must be one of float, vec3, mat3, quat, mat4')
	
	slvvars = ('origin', 'direction')
	def slv_tangent(self, pt):
		return self.direction
		
	def __repr__(self):
		return 'Axis({}, {})'.format(self.origin, self.direction)
	
	def display(self, scene):
		return displays.AxisDisplay(scene, (self.origin, self.direction), self.interval)
			
def isaxis(obj):
	''' return True if the given object is considered to be an axis.
		An axis can be an instance of `Axis` or a tuple `(vec3, vec3)`
	'''
	return isinstance(obj, Axis) or isinstance(obj, tuple) and len(obj)==2 and isinstance(obj[0],vec3) and isinstance(obj[1],vec3)


class Segment(object):
	''' segment from a to b '''
	__slots__ = ('a', 'b')
	def __init__(self, a, b):
		self.a, self.b = a,b
		
	@property
	def direction(self):
		return normalize(self.b-self.a)
		
	@property
	def origin(self):
		return mix(self.a, self.b, 0.5)
		
	slvvars = 'a', 'b'
	def slv_tangent(self, pt):
		return self.direction
	
	def mesh(self):
		return mesh.Wire([self.a, self.b], groups=[None])
		
	def __repr__(self):
		return 'Segment({}, {})'.format(self.a, self.b)
	
	def display(self, scene):	
		return self.mesh().display(scene)

class ArcThrough(object):	
	''' arc from a to c, passing through b '''
	__slots__ = ('a', 'b', 'c', 'resolution')
	
	def __init__(self, a,b,c, resolution=None):
		self.a, self.b, self.c = a,b,c
		self.resolution = resolution
	
	@property
	def center(self):
		a,c = self.a-self.b, self.c-self.b
		l1,l2 = length(a),length(c)
		t = dot(a,c)/(l1*l2)
		h = 0.5 * (l1/l2-t) * 1/sqrt(1-t**2)
		n = normalize(cross(self.a-self.b, self.c-self.b))
		return (self.c+self.b)/2 + cross(n, self.b-self.c) * h
	
	@property
	def radius(self):
		return length(self.a-self.center)
	
	@property
	def axis(self):
		return (self.center, normalize(cross(self.a-self.b, self.c-self.b)))
	
	def tangent(self, pt):
		''' tangent to the closest point of the curve to pt '''
		c = self.center
		return normalize(cross(pt-c, cross(self.a-c, self.c-c)))
	
	slvvars = ('a', 'b', 'c')
	slv_tangent = tangent
	
	def mesh(self, resolution=None):
		center = self.center
		z = normalize(cross(self.c-self.b, self.a-self.b))
		return mkarc((center, z), self.a, self.c, resolution or self.resolution)
		
	def __repr__(self):
		return 'Axis({}, {}, {})'.format(self.a, self.b, self.c)
	
	def display(self, scene):	
		return self.mesh().display(scene)

class ArcCentered(object):
	''' arc from a to b, centered around the origin of the axis.
	
		An axis is requested instead of a point (that would be more intuitive), to solve the problem when a,b, center are aligned
	'''
	__slots__ = ('axis', 'a', 'b', 'resolution')
	def __init__(self, axis, a, b, resolution=None):
		self.axis = axis
		self.a, self.b = a, b
		self.resolution = resolution
	
	@property
	def center(self):
		return self.axis[0]
	
	@property
	def radius(self):
		return (distance(self.axis[0], self.a) + distance(self.axis[0], self.b)) /2
	
	def tangent(self, pt):
		''' tangent to the closest point of the curve to pt '''
		return cross(normalize(pt - self.axis[0]), self.axis[1])
	
	slvvars = ('axis', 'a', 'b')
	slv_tangent = tangent
	def fit(self):
		return (dot(self.a-self.axis[0], self.axis[1]) **2,
				dot(self.b-self.axis[0], self.axis[1]) **2,
				(distance(self.a,self.axis[0]) - distance(self.b,self.axis[0])) **2,
				(length(self.axis[1])-1) **2,
				)
	
	def mesh(self, resolution=None):
		return mkarc(self.axis, self.a, self.b, resolution or self.resolution)
		
	def __repr__(self):
		return 'ArcCentered({}, {}, {})'.format(self.axis, self.a, self.b)
	
	def display(self, scene):	
		return self.mesh().display(scene)
		
class ArcTangent(object):
	''' An arc always tangent to `Segment(a,b)` and `Segment(c,b)`. The solution is unique.'''
	__slots = 'a', 'b', 'c'
	def __init__(self, a, b, c, resolution=None):
		self.a, self.b, self.c = a,b,c
		self.resolution = resolution
	
	@property
	def center(self):
		n = normalize(self.a + self.c - 2*self.b)
		return mix(	unproject(self.a-self.b, n),
					unproject(self.c-self.b, n),
					0.5)	+ self.b
	
	@property
	def radius(self):
		c = self.center
		return mix(distance(c,self.a), distance(c,self.c), 0.5)
	
	@property
	def axis(self):
		return self.center, normalize(cross(self.b-self.a, self.c-self.a))
	
	def tangent(self, pt):
		''' tangent to the closest point of the curve to pt '''
		z = cross(self.a-self.b, self.c-self.b)
		return normalize(cross(pt-self.center, z))
	
	slvvars = 'a', 'b', 'c'
	slv_tangent = tangent
	
	def fit(self):
		return (distance(self.a, self.b) - distance(self.c, self.b)) **2,
	
	def mesh(self, resolution=None):
		return mkarc(self.axis, self.a, self.c, resolution or self.resolution)
		
	def __repr__(self):
		return 'ArcTangent({}, {}, {})'.format(self.a, self.b, self.c)
	
	def display(self, scene):	
		return self.mesh().display(scene)

def mkarc(axis, start, end, resolution=None):
	center, z = axis
	center = center + dot(z, (start+end)/2-center) * z
	v = start-center
	r = length(v)
	x = v/r
	y = cross(z, x)
	angle = atan2(dot(end-center,y), dot(end-center,x)) % (2*pi)
	div = settings.curve_resolution(angle*r, angle, resolution)
	pts = [start]
	for i in range(1,div+1):
		a = angle * i/(div+1)
		pts.append(x*r*cos(a) + y*r*sin(a) + center)
	pts.append(end)
	return mesh.Wire(pts, groups=[None])

class TangentEllipsis(object):
	''' An quater of ellipsis always tangent to `Segment(a,b)` and `Segment(c,b)`. The solution is unique.
	'''
	__slots__ = ('a', 'b', 'c', 'resolution')
	def __init__(self, a,b,c, resolution=None):
		self.a, self.b, self.c = a,b,c
		self.resolution = resolution
	
	@property
	def axis(self):
		return (self.center, normalize(cross(self.a-self.b, self.c-self.b)))
		
	@property
	def center(self):
		return self.a + self.c - self.b
	
	def tangent(self, pt):
		''' tangent to the closest point of the curve to pt '''
		c,z = self.axis
		return normalize(cross(z, pt - c))
	
	slvvars = ('a', 'b', 'c')
	slv_tangent = tangent
	
	def mesh(self, resolution=None):
		''' axis directions doesn't need to be normalized nor oriented '''
		origin = self.b
		x = origin - self.a
		y = origin - self.c
		div = settings.curve_resolution(distance(self.a,self.c), anglebt(x,-y), self.resolution or resolution)
		pts = [self.a]
		for i in range(div+1):
			t = pi/2 * i/(div+1)
			pts.append(x*sin(t) + y*cos(t) + origin-x-y)
		pts.append(self.c)
		return mesh.Wire(pts, groups=[None])
		
	def __repr__(self):
		return 'TangentEllipsis({}, {}, {})'.format(self.c, self.b, self.c)
	
	def display(self, scene):	
		return self.mesh().display(scene)

class Circle(object):
	''' circle centered around the axis origin, with the given radius, in an orthogonal plane to the axis direction '''
	__slots__ = ('axis', 'radius', 'alignment', 'resolution')
	def __init__(self, axis, radius, alignment=vec3(1,0,0), resolution=None):
		self.axis, self.radius = axis, radius
		self.alignment = alignment
		self.resolution = resolution
	
	@property
	def center(self):
		return self.axis[0]
	
	def fit(self):
		return (length(self.axis[1])-1) **2,
	
	def tangent(self, pt):
		''' tangent to the closest point of the curve to pt '''
		return normalize(cross(pt-self.axis[0], self.axis[1]))
	
	slvvars = ('axis', 'radius')
	slv_tangent = tangent
	
	def mesh(self, resolution=None):
		center = self.axis[0]
		x,y,z = dirbase(self.axis[1], self.alignment)
		angle = 2*pi
		r = self.radius
		div = settings.curve_resolution(angle*r, angle, self.resolution or resolution)
		pts = []
		for i in range(div):
			a = angle * i/div
			pts.append(x*r*cos(a) + y*r*sin(a) + center)
		indices = list(range(div))
		indices.append(0)
		return mesh.Wire(pts, indices, groups=[None])
		
	def __repr__(self):
		return 'Circle({}, {})'.format(self.axis, self.radius)
	
	def display(self, scene):	
		return self.mesh().display(scene)
		
		
import numpy.core as np
def glmarray(array, dtype='f4'):
	''' create a numpy array from a list of glm vec '''
	buff = np.empty((len(array), len(array[0])), dtype=dtype)
	for i,e in enumerate(array):
		buff[i][:] = e
	return buff

		
class Interpolated(object):
	''' interpolated curve passing through the given points (3rd degree bezier spline) 

		the tangent in each point is determined by the direction between adjacent points
		the point weights is how flattened is the curve close to the point tangents
	'''
	__slots__ = 'points', 'weights', 'resolution'
	def __init__(self, points, weights=None, resolution=None):
		self.points = points
		self.weights = weights or [1] * len(self.points)
		self.resolution = resolution
		
	def mesh(self, resolution=None):
		pts = self.points
		if not pts:		return Wire()
		
		# get tangent to each point
		tas = [self.weights[i-1] * length(pts[i]-pts[i-1]) * normalize(pts[i]-pts[i-2])	
					for i in range(2,len(pts))]
		tbs = [self.weights[i-1] * length(pts[i-2]-pts[i-1]) * normalize(pts[i-2]-pts[i])	
					for i in range(2,len(pts))]
		tas.insert(0, tbs[0] - 2*project(tbs[0], pts[1]-pts[0]))
		tbs.append(tas[-1] - 2*project(tas[-1], pts[-2]-pts[-1]))
		
		# stack points to curve
		curve = []
		for i in range(len(pts)-1):
			a,b = pts[i], pts[i+1]
			ta,tb = tas[i], tbs[i]
			# tangent to the curve in its inflexion point
			mid = 1.25*(b-a) + 0.25*(tb-ta)
			# get resolution
			div = 1 + settings.curve_resolution(
							length(ta) + length(tb), 
							anglebt(-tb, mid) + anglebt(ta, mid),
							self.resolution or resolution)
			
			# append the points for this segment
			for i in range(div+1):
				curve.append(interpol2((a,ta), (b,tb), i/(div+1)))
		
		curve.append(b)
		return mesh.Wire(curve, groups=[None])
		
	def box(self):
		return boundingbox(self.points)
		
	def __repr__(self):
		return 'Interpolated({}, {})'.format(self.points, self.weights)

	def display(self, scene):
		return displays.SplineDisplay(scene, glmarray(self.points), glmarray(self.mesh().points))
			
class Softened(object):
	''' interpolated curve tangent to each segment midpoint (3rd degree bezier curve)
	
		the points weights is the weight in the determination of each midpoint
	'''
	__slots__ = 'points', 'weights', 'resolution'
	def __init__(self, points, weights=None, resolution=None):
		self.points = points
		self.weights = weights or [1] * len(self.points)
		self.resolution = resolution
		
	def mesh(self, resolution=None):
		pts = self.points
		if not pts:		return Wire()
		
		# find midpoints
		mid = [	(self.weights[i-1]*pts[i-1] + self.weights[i]*pts[i]) 
					/ (self.weights[i-1]+self.weights[i])
				for i in range(1,len(pts)) ]
		mid[0] = pts[0]
		mid[-1] = pts[-1]
		
		# stack points to curve
		curve = []
		for i in range(len(pts)-2):
			a,b = mid[i],  mid[i+1]
			ta,tb = 2*(pts[i+1]-a),  2*(pts[i+1]-b)
			# get resolution
			div = 1 + settings.curve_resolution(
						length(ta) + length(tb), 
						anglebt(ta, -tb),
						self.resolution or resolution)
			
			# append the points for this segment
			for i in range(div+1):
				curve.append(interpol2((a,ta), (b,tb), i/(div+1)))
		
		curve.append(b)
		return mesh.Wire(curve, groups=[None])
		
	def box(self):
		return boundingbox(self.points)
		
	def __repr__(self):
		return 'Softened({}, {})'.format(self.points, self.weights)

	def display(self, scene):
		return displays.SplineDisplay(scene, glmarray(self.points), glmarray(self.mesh().points))
	



# This file is part of pymadcad,  distributed under license LGPL v3

from math import inf
from .mathutils import *
from .mesh import Mesh, Web, Wire, MeshError, connpe, web
from .asso import Asso
from .nprint import nformat, nprint

from copy import copy
from operator import itemgetter


class TriangulationError(Exception):	pass


def triangulation(outline, normal=None, prec=NUMPREC):
	''' triangulate using the prefered method '''
	return triangulation_closest(outline, normal)
	
	
	

'''
	skeleting funtions and skeleting triangulations
	does not work yet on very non convex cases
'''

def skeleting(outline: Wire, skeleting: callable, prec=NUMPREC) -> [vec2]:
	''' skeleting procedure for the given wire
		at each step, the skeleting function is called
		created points will be added to the wire point buffer and this buffer is returned (ROI)
		
		NOTE: yet it only works with near-convex outlines
	'''
	l = len(outline)
	pts = outline.points
	
	# edge normals
	enormals = [perp(normalize(outline[i]-outline[i-1]))	for i in range(l)]
	# compute half axis starting from each point
	haxis = []
	for i in range(l):
		haxis.append((outline.indices[i], i, (i+1)%l))
	
	# create the intersections to update
	intersect = [(0,0)] * l	# intersection for each edge
	dist = [-1] * l	# min distance for each edge
	def eval_intersect(i):
		o1,a1,b1 = haxis[i-1]
		o2,a2,b2 = haxis[i]
		# compute normals to points
		v1 = enormals[a1]+enormals[b1]
		v2 = enormals[a2]+enormals[b2]
		if norminf(v1) < prec:	v1 =  perp(enormals[b1])	# if edges are parallel, take the direction to the shape
		if norminf(v2) < prec:	v2 = -perp(enormals[b2])
		# compute the intersection
		x1,x2 = inverse(dmat2(v1,-v2)) * dvec2(-pts[o1] + pts[o2])
		if x1 >= -NUMPREC and x2 >= -NUMPREC:
			intersect[i] = pts[o2] + x2*v2
			dist[i] = min(x1*length(v1), x2*length(v2))
		elif isnan(x1) or isnan(x2):
			intersect[i] = pts[o2]
			dist[i] = 0
		else:
			intersect[i] = None
			dist[i] = inf
	for i in range(l):
		eval_intersect(i)
	
	# build skeleton
	while len(haxis) > 1:
		print(dist)
		i = min(range(len(haxis)), key=lambda i:dist[i])
		assert dist[i] != inf, "no more intersection found (algorithm failure)"
		o1,a1,b1 = haxis[i-1]
		o2,a2,b2 = haxis[i]
		# add the intersection point
		ip = len(pts)
		pts.append(intersect[i])
		# extend skeleton
		skeleting(haxis, i, ip)
		# create the new half axis
		haxis.pop(i)
		dist.pop(i)
		intersect.pop(i)
		haxis[i-1] = (ip, a1, b2)
		eval_intersect((i-2) % len(haxis))
		eval_intersect((i-1) % len(haxis))
		eval_intersect(i % len(haxis))
		eval_intersect((i+1) % len(haxis))
		eval_intersect((i+2) % len(haxis))
	
	return pts

def skeleton(outline: Wire, prec=NUMPREC) -> Web:
	''' return a Web that constitute the skeleton of the outline
		the returned Web uses the same point buffer than the input Wire.
		created points will be added into it
		
		https://en.wikipedia.org/wiki/Straight_skeleton
		
		NOTE: yet it only works with near-convex outlines
	'''
	skeleton = []
	#def sk(ip, o1,o2, a1,b1, a2,b2):
		#skeleton.append((o1,ip))
		#skeleton.append((o2,ip))
	def sk(haxis, i, ip):
		skeleton.append((haxis[i-1][0], ip))
		skeleton.append((haxis[i][0], ip))
	pts = skeleting(outline, sk, prec)
	return Web(pts, skeleton)

def triangulation_skeleton(outline: Wire, prec=NUMPREC) -> Mesh:
	''' return a Mesh with triangles filling the surface of the outline 
		the returned Mesh uses the same point buffer than the input Wire
		
		NOTE: yet it only works with near-convex outlines
	'''
	triangles = []
	skeleton = []
	pts = outline.points
	original = len(pts)
	minbone = [inf]
	def sk(haxis, i, ip):
		triangles.append((haxis[i-2][0], haxis[i-1][0], ip))
		triangles.append((haxis[i-1][0], haxis[i][0], ip))
		triangles.append((haxis[i][0], haxis[(i+1)%len(haxis)][0], ip))
		if   haxis[i-1][0] < original:
			d = distance(pts[haxis[i-1][0]], pts[ip])
			if d < minbone[0]:	minbone[0] = d
		else:
			skeleton.append((haxis[i-1][0], ip))
		if haxis[i][0] < original:
			d = distance(pts[haxis[i][0]], pts[ip])
			if d < minbone[0]:	minbone[0] = d
		else:
			skeleton.append((haxis[i][0], ip))
	pts = skeleting(outline, sk, prec)
	m = Mesh(pts, triangles)
	# merge points from short internal edges
	minbone = 0.5*minbone[0]
	merges = {}
	for a,b in skeleton:
		if distance(pts[a], pts[b]) < minbone:
			if   a not in merges:	merges[a] = merges.get(b,b)
			elif b not in merges:	merges[b] = merges.get(a,a)
	for k,v in merges.items():
		while v in merges and merges[v] != v:	
			if distance(pts[k], pts[v]) > minbone:
				merges[k] = k
				merges[v] = v
				break
			merges[k] = v = merges[v]
	m.mergepoints(merges)
	return m


	
'''
	triangulation outline
	triangulation using only points from the given outline, creating quite good triangles at cost of efficiency
	O(n**2)
'''

def triangulation_outline(outline: Wire, normal=None) -> Mesh:
	''' return a mesh with the triangles formed in the outline
		the returned mesh uses the same buffer of points than the input
		
		complexity:  O(n**2) mat2 operations
	'''
	# get a normal in the right direction for loop winding
	if not normal:		normal = outline.normal()
	try:				proj = planeproject(outline, normal)
	except ValueError:	return Mesh()
	hole = list(outline.indices)
	if length2(outline[-1]-outline[0]) < 4*NUMPREC**2:		hole.pop()
	
	def priority(u,v):
		''' priority criterion for 2D triangles, depending on its shape
		'''
		uv = length(u)*length(v)
		if not uv:	return 0
		return dot(u,v) / uv
	
	def score(i):
		l = len(hole)
		u = proj[(i+1)%l] - proj[i]
		v = proj[(i-1)%l] - proj[i]
		triangle = (hole[(i-1)%l], hole[i], hole[(i+1)%l])
		
		if triangle[0] == triangle[1] or triangle[1] == triangle[2]:	return 0
		if perpdot(u,v) < -NUMPREC:		return -inf
		sc = priority(u,v)
		if perpdot(u,v) > NUMPREC:
			# check that there is not point of the outline inside the triangle
			decomp = inverse(dmat2(u,v))
			o = proj[i]
			for j in range(l):
				if hole[j] not in triangle:
					p = proj[j]
					uc,vc = decomp * dvec2(p-o)
					if 0 <= uc and 0 <= vc and uc+vc <= 1:
						sc = -inf
						break
		return sc
	scores = [score(i) for i in range(len(hole))]
	
	triangles = []
	while len(hole) > 2:
		l = len(hole)
		i = max(range(l), key=lambda i:scores[i])
		if scores[i] == -inf:
			raise TriangulationError("no more feasible triangles (algorithm failure or bad input outline)", hole)
			#print('warning: no more feasible triangles', hole)
			#break
		triangles.append((hole[(i-1)%l], hole[i], hole[(i+1)%l]))
		hole.pop(i)
		proj.pop(i)
		scores.pop(i)
		l -= 1
		scores[(i-1)%l] = score((i-1)%l)
		scores[i%l] = score(i%l)
	
	return Mesh(outline.points, triangles)


def triangulation_outline(outline: Wire, normal=None, prec=None) -> Mesh:
	''' return a mesh with the triangles formed in the outline
		the returned mesh uses the same buffer of points than the input
		
		complexity:  O(n*k)  where k is the number of non convex points
	'''
	# get a normal in the right direction for loop winding
	if not normal:		normal = outline.normal()
	if prec is None:	prec = outline.precision()
	# the precision we will use in this function is a surface precision:  maximum edge length * minimum edge length
	prec = prec**2/NUMPREC 
	# project all points in the plane
	try:				proj = planeproject(outline, normal)
	except ValueError:	return Mesh()
	# reducing contour, indexing proj and outlines.indices
	hole = list(range(len(outline.indices)))
	if length2(outline[-1]-outline[0]) <= prec:		hole.pop()
	
	# set of remaining non-convexity points, indexing proj
	l = len(outline.indices)
	nonconvex = { i
					for i in hole
					if perpdot(proj[i]-proj[i-1], proj[(i+1)%l]-proj[i]) <= prec
					}
	
	def priority(u,v):
		''' priority criterion for 2D triangles, depending on its shape
		'''
		uv = length(u)*length(v)
		if not uv:	return 0
		return dot(u,v) / uv
	
	def score(i):
		l = len(hole)
		o = proj[hole[ i ]]
		u = proj[hole[ (i+1)%l ]] - o
		v = proj[hole[ (i-1)%l ]] - o
		triangle = (hole[(i-1)%l], hole[i], hole[(i+1)%l])
		
		# check for badly oriented triangle
		if perpdot(u,v) < -prec:		
			return -inf
		# check for intersection with the rest
		elif perpdot(u,v) > prec:
			# check that there is not point of the outline inside the triangle
			for j in nonconvex:
				if j not in triangle:
					for k in range(3):
						a,b = proj[triangle[k]], proj[triangle[k-1]]
						s = perpdot(a-b, proj[j]-a)
						if k == 1:	s -= 2*prec
						if s <= -prec:
							break
					else:
						return -inf
		# flat triangles that change direction are forbidden, it must be considered as a +360° angle and not a 0° angle
		elif dot(u,v) >= prec:
			return -inf
		return priority(u,v)
	scores = [score(i) for i in range(len(hole))]
	
	triangles = typedlist(dtype=uvec3)
	while len(hole) > 2:
		l = len(hole)
		i = imax(scores)
		if scores[i] == -inf:
			raise TriangulationError("no more feasible triangles (algorithm failure or bad input outline)", [outline.indices[i] for i in hole])
			#print('warning: no more feasible triangles', [outline.indices[i] for i in hole])
			#a, b, c, *_ = hole
			#print('area', prec, perpdot(proj[a]-proj[b], proj[c]-proj[b]), dot(proj[a]-proj[b], proj[c]-proj[b]))
		
		triangles.append(uvec3(
			outline.indices[hole[(i-1)%l]], 
			outline.indices[hole[i]], 
			outline.indices[hole[(i+1)%l]],
			))
		nonconvex.discard(hole.pop(i))
		scores.pop(i)
		l -= 1
		scores[(i-1)%l] = score((i-1)%l)
		scores[i%l] = score(i%l)
	
	return Mesh(outline.points, triangles)

def planeproject(pts, normal=None):
	''' project an outline in a plane, to get its points as vec2 '''
	x,y,z = guessbase(pts, normal)
	return typedlist(vec2(dot(p,x), dot(p,y))	for p in pts)

def guessbase(pts, normal=None, thres=10*NUMPREC):
	''' build a base in which the points will be in plane XY 
		thres is the precision threshold between axis for point selection
	'''
	if normal:
		return dirbase(normal)
	pts = iter(pts)
	try:
		o = next(pts)
		x = vec3(0)
		y = vec3(0)
		ol = max(glm.abs(o))
		xl = 0
		zl = 0
		while xl <= thres:
			p = next(pts)
			x = p-o
			xl = dot(x,x)/max(max(glm.abs(p)), ol)
		x = normalize(x)
		while zl <= thres:
			p = next(pts)
			y = p-o
			zl = length(cross(x,y))
		y = normalize(noproject(y,x))
		return x, y, cross(x,y)
	except StopIteration:
		raise ValueError('unable to extract 2 directions')

def convex_normal(web):
	area = vec3(0)
	pts = web.points
	c = pts[web.edges[0][0]]
	for e in web.edges:
		area += cross(pts[e[1]] - pts[e[0]],  c - pts[e[0]])
	return normalize(area)
	
	
'''
	triangulation_sweepline
	---------------------
	loop holes briding based on a sweepline algorithm - complexity O(n log n)
'''

def sweepline_loops(lines: Web, normal=None):
	''' sweep line algorithm to retreive monotone loops from a Web
		the web edges should not be oriented, and thus the resulting face has no orientation
		complexity: O(n ln n)
		
		https://en.wikipedia.org/wiki/Monotone_polygon
	'''
	# selective debug function
	#debprint = lambda *args, **kwargs: None
	#debprint = print

	if len(lines.edges) < 3:	return Mesh()
	loops = []
	finalized = []
	x,y,z = guessbase(lines.points, normal)
	#debprint('sortdim', x, 'y', y)
	# projection of the lines points on x and y
	pts = {}
	for e in lines.edges:
		for i in e:
			if i not in pts:
				p = lines.points[i]
				pts[i] = vec2(dot(p,x), dot(p,y))
	
	#debprint('sweepline')
	#debprint('\npoints = {},\nedges = {},\n'.format(lines.points, lines.edges))
	
	# affine function y = a*x + b   for edges
	def affiney(e, x):
		a, b = pts[e[0]], pts[e[1]]
		v = b-a
		if v[0]:	d = v[1]/v[0]
		elif v[1] > 0:	d = -inf
		else:			d = inf
		return a[1] + d * (x - a[0])
	
	def orthoproj(v):
		l = length(v)
		return v[1] / l if l else 0
	
	# orient edges along the axis and sort them
	# sorting is done using absciss and orientation in case of similar absciss
	# the nearest directions to +-y is prefered to speedup cluster distinction
	edges = lines.edges[:]
	for i,(a,b) in enumerate(edges):
		if pts[a][0] < pts[b][0]:	edges[i] = b,a
	stack = sorted(edges,
				key=lambda e: (pts[e[0]][0], abs(orthoproj(pts[e[1]]-pts[e[0]])) )
				)
	
	# remove absciss ambiguity (for edges that share points with the same absciss)
	# for each edge, all following edges with the same absciss and contains its start point will have the same startpoint
	for i in reversed(range(len(stack))):
		l = stack[i][0]
		n = i-1
		for j in reversed(range(i)):
			if pts[stack[j][0]][0] < pts[l][0]:	break
			if stack[j][1] == l:	
				stack[j] = (stack[j][1], stack[j][0])
			if stack[j][0] == l:
				if n != j:	stack.insert(n, stack.pop(j))
				n -= 1
	#debprint('stack', stack)
	#debprint('stack')
	#for e in stack:
		#debprint(e, (pts[e[0]][0], orthoproj(pts[e[1]]-pts[e[0]])))
	
	# build cluster by cluster -  each cluster is a monotone sub-polygon
	# kept sorted so that:
	#    clusters are in descending order on y
	# 	 cluster couple is in ascending order:  (l0, l1) with l0 < l1
	clusters = []
	while stack:		
		edge = stack.pop()
		p0, p1 = pts[edge[0]], pts[edge[1]]
		
		# get the pair edge if there is one starting at the same point
		m = None
		sc = -1
		i = len(stack)-1
		while i >= 0 and pts[stack[i][0]][0] == pts[edge[0]][0]:
			e = stack[i]
			if e[0] == edge[0]:
				diff = abs(orthoproj(pts[e[1]]-pts[e[0]]) - orthoproj(pts[edge[1]]-pts[edge[0]]))
				if diff > sc:	
					sc = diff
					m = i
			i -= 1
		if m is not None:	
			# edge must be above coedge
			coedge = stack.pop(m)
			if orthoproj(pts[edge[1]]-pts[edge[0]]) < orthoproj(pts[coedge[1]]-pts[coedge[0]]):
				edge, coedge = coedge, edge
		else:
			coedge = None
			
		# finalize closed clusters, merge clusters that reach the same points
		for i,(l0,l1) in enumerate(clusters):	
			# remove the closed cluster
			if l0[1] == l1[1]:
				#debprint('    pop', l0,l1)
				clusters.pop(i)
				loops[i].insert(0, l0[1])
				finalized.append(loops.pop(i))
				break
			# merge the two neighboring clusters
			if i>0 and clusters[i-1][0][1] == l1[1] and pts[l1[1]][0] > p0[0]:
				merge = clusters.pop(i)
				clusters[i-1] = (merge[0], clusters[i-1][1])
				loops[i].append(l1[1])
				loops[i-1][0:0] = loops.pop(i)
				#debprint('    merged', loops[i-1])
				#debprint('clusters', nformat(clusters))
				break
		
		#print('stack', stack)
		#debprint('*', edge, coedge, pts[edge[0]][0], -abs(orthoproj(pts[edge[1]]-pts[edge[0]])))
		
		# search in which cluster we are
		found = False
		for i,(l0,l1) in enumerate(clusters):
			#debprint('   ', l0,l1,edge[0])
			
			# continuation of already existing edge of the cluster
			if edge[0] == l0[1]:
				loops[i].insert(0, l0[1])
				clusters[i] = (edge, l1)
				if coedge:
					stack.append(coedge)
				#debprint('      continuation', edge)
				#if coedge:	nprint(clusters)
				found = True
				break
			elif edge[0] == l1[1]:
				loops[i].append(l1[1])
				clusters[i] = (l0, coedge or edge)
				if coedge:
					stack.append(edge)
				#debprint('      continuation', coedge or edge)
				#if coedge:	nprint(clusters)
				found = True
				break
			# interior hole that touch the outline
			elif (coedge and l0[0] == l1[0] and l0[0] == edge[0]
					and affiney(l0, p1[0]) <= p1[1]
					and affiney(l1, p1[0]) >= p1[1]):
				clusters[i] = (l0, coedge)
				clusters.insert(i, (edge, l1))
				loops.insert(i, [edge[0]])
				#debprint('      root hole', edge[0])
				found = True
				break
			# interior hole
			elif (coedge
					and affiney(l0, p0[0]) < p0[1]
					and affiney(l1, p0[0]) > p0[1]):
				clusters[i] = (l0, coedge)
				clusters.insert(i, (edge, l1))
				# continue the cluster on the side with the higher x
				if pts[l0[0]][0] < pts[l1[0]][0]:
					# the existing loop stays at index i
					loops[i].insert(0,edge[0])
					loops.insert(i+1, [l0[0], coedge[0]])
				else:
					# the existing loop jumps to index i+1
					loops[i].append(coedge[0])
					loops.insert(i, [edge[0], l1[0]])
				#debprint('      hole for ',i, edge)
				#debprint('clusters', nformat(clusters))
				#debprint('loops', loops)
				found = True
				break
		
		if not found:
			# if it's a new corner, create a cluster
			if coedge and edge[1] != coedge[1]:
				p0 = pts[edge[0]]
				# find the place to insert the cluster
				# NOTE a dichotomy is more efficient, but for now ...
				j = 0
				while j < len(clusters) and affiney(clusters[j][1], p0[0]) > p0[1]:
					j += 1
				clusters.insert(j, (coedge, edge))
				loops.insert(j, [edge[0]])
				#debprint('    new cluster', j)
				#debprint(nformat(clusters))
			# if it's an ambiguous edge without use for now, restack it
			elif pts[edge[1]][0] == pts[edge[0]][0]:
				#debprint('    restack')
				stack.insert(-1, edge)
				if coedge:	stack.insert(-1, coedge)
			else:
				raise TriangulationError("algorithm failure, can be due to the given outline", edge, coedge)
	
	# close clusters' ends
	for i,cluster in enumerate(clusters):
		l0,l1 = cluster
		loops[i].insert(0, l0[1])
		if l0[1] != l1[1]:
			loops[i].append(l1[1])
	finalized.extend(loops)
	#for loop in finalized:
		#debprint('triangulate', loop)
	
	return finalized


def triangulation_sweepline(outline: Web, normal=None, prec=0) -> Mesh:
	''' extract loops from the web using sweepline_loops and trianglate them.
		the resulting mesh have one group per loop, and all the normals have approximately the same direction (they are coherent)
	'''
	x,y,z = guessbase(outline.points, normal)
	m = Mesh(outline.points)
	for loop in sweepline_loops(outline, z):
		m += triangulation_outline(Wire(outline.points, loop), z)
	return m
	

def line_bridges(lines: Web, conn=None) -> 'Web':
	''' find what edges to insert in the given mesh to make all its loops connex.
		returns a Web of the bridging edges.
		
		complexity:  O(n**2 + n*k)
		with
			n = number of points in the lines
			k = average number of points per loop
	'''
	if conn is None:	conn = connpe(lines.edges)
	pts = lines.points
	
	bridges = []	# edges created
	# reached and remaining parts, both as edge indices and point indices
	reached_points = {}	# closest to each reached point
	reached_edges = {}	# closest to each reached edge
	remain_points = set(p 	 for e in lines.edges for p in e)
	remain_edges = set(range(len(lines.edges)))
	
	def propagate(start):
		''' propagate from the given start point to make connex points and edges as reached '''
		front = [start]
		while front:
			s = front.pop()
			if s in reached_points:	continue
			reached_points[s] = (inf, (s,s))
			remain_points.discard(s)
			for e in conn[s]:
				if e in reached_edges: continue
				reached_edges[e] = (inf, (s,s))
				remain_edges.discard(e)
				front.extend(lines.edges[e])
				
	def update_closest():
		# from points to edges
		for s, (score, best) in reached_points.items():
			if best[0] in reached_points:
				reached_points[s] = find_closest_point(s)
		# from edges to points
		for e, (score, best) in reached_edges.items():
			if best[0] in reached_points:
				reached_edges[e] = find_closest_edge(e)
	
	def find_closest_edge(ac):
		''' find the bridge to the closest point to the given edge
		'''
		best = None
		score = inf
		# minimum from edge to points
		ep = lines.edgepoints(ac)
		for ma in remain_points:
			d = distance_pe(pts[ma], ep)
			if d < score:
				e = lines.edges[ac]
				if distance2(pts[ma], pts[e[0]]) < distance2(pts[ma], pts[e[1]]):
					score, best = d, (ma, e[0])
				else:
					score, best = d, (ma, e[1])
		
		return score, best
		
	def find_closest_point(ac):
		''' find the bridge to the closest edge to the given point
		'''
		best = None
		score = inf
		# minimum from point to edges
		for ma in remain_edges:
			d = distance_pe(pts[ac], lines.edgepoints(ma))
			if d < score:
				e = lines.edges[ma]
				if distance2(pts[ac], pts[e[0]]) < distance2(pts[ac], pts[e[1]]):
					score, best = d, (e[0], ac)
				else:
					score, best = d, (e[1], ac)
						
		return score, best
	
	# main loop
	propagate(lines.edges[0][0])
	while remain_edges:
		update_closest()
		closest = min(
					min(reached_points.values(), key=itemgetter(0)),
					min(reached_edges.values(), key=itemgetter(0)),
					key=itemgetter(0)) [1]
		bridges.append(closest)
		bridges.append(tuple(reversed(closest)))
		propagate(closest[0])
		
	return Web(lines.points, bridges)
	

	
	
	
def flat_loops(lines: Web, normal=None) -> '[Wire]':
	''' collect the closed loops present in the given web, so that no loop overlap on the other ones.
	'''
	x,y,z = guessbase(lines.points, normal)
	pts = lines.points
	# create an oriented point to edge connectivity 
	conn = Asso()
	for i,e in enumerate(lines.edges):
		conn.add(e[0], i)
	
	used = [False]*len(lines.edges)	# edges used and so no to use anymore	
	# the start point for loop assembly must be in priority on an edge that has no double, because double edges that are not on a bridge would create independent loops out of an empty surface
	# create lists of edges that have a double and edges that have not
	doubled = []
	nondoubled = []
	for i, e in enumerate(lines.edges):
		double = False
		for j in conn[e[1]]:
			if lines.edges[j][1] == e[0]:
				double = True
		if double:	doubled.append(i)
		else:		nondoubled.append(i)
	# iterator of start points
	choice = (i  for i in (nondoubled + doubled)  if not used[i])
	
	loops = []
	
	while True:
		# find an unused edge
		end = next(choice, None)
		if end is None:
			break
		
		# assemble a loop
		loop = list(lines.edges[end])
		while True:
			# take the most inward next edge
			prev = normalize(pts[loop[-1]] - pts[loop[-2]])
			
			best = None
			score = -inf
			for edge in conn[loop[-1]]:
				if used[edge]:	continue
				e = lines.edges[edge]
				dir = normalize(pts[e[1]] - pts[e[0]])
				if isfinite(dir) and isfinite(prev):
					angle = atan2(dot(cross(prev,dir),z), dot(prev,dir))
				else:
					angle = -pi
				if pi-angle <= pi*NUMPREC:	angle -= 2*pi
				if angle > score:
					score, best = angle, edge
					
			
			# progress on the selected edge
			if best is None:
				raise TriangulationError("there is not only loops in that web", loop)
			used[best] = True
			if best == end:
				break
			
			# continue the loops
			loop.append(lines.edges[best][1])
			
			# the very particular case of loops with null surface, or with straight return to origin might intruduce serveral loops in the same suite, the following condition prevents it
			if loop[0] == loop[-1]:
				prev = normalize(pts[loop[-2]] - pts[loop[-1]])
				dir = normalize(pts[loop[1]] - pts[loop[0]])
				if acos(dot(prev,dir)) <= pi*NUMPREC:
					used[end] = True
					break
			
		loops.append(Wire(lines.points, loop))

	return loops

	
def triangulation_closest(outline, normal=None, prec=None):
	if isinstance(outline, Wire):
		return triangulation_outline(outline, normal, prec)
	else:
		outline = copy(web(outline)).mergegroups()
	x,y,z = dirbase(normal or convex_normal(outline))
	result = Mesh(outline.points)
	for loop in flat_loops(outline + line_bridges(outline), z):
		result += triangulation_outline(loop, z, prec)
	return result

	
	
	
	
	
	
''' 
	other misc stuff, in development
'''

	
from .mesh import connef, arrangeface

def retriangulate(mesh):
	''' switch diagonals to improve the surface smoothness '''
	pts = mesh.points
	faces = mesh.faces
	# second pass:	improve triangles by switching quand diagonals
	conn = connef(faces)
	scores = {}
	def update_score(edge):
		a,b,c = arrangeface(faces[conn[edge]], edge[0])
		ab = pts[b]-pts[a]
		lab = length(ab)
		# check if diagonal is good for exchange
		x = dot(pts[c]-pts[a], ab) / (lab*lab) 
		if x < 0 or 1 < x:
			scores[edge] = 0
		else:
			# if then, record the difference of heights
			h = length(noproject(pts[c]-pts[a], ab/lab))
			scores[edge] = lab/h - 2
	for edge in conn:
		update_score(edge)
	
	while True:
		a,b = min(scores, key=lambda e:scores[e])
		if scores[edge] <= NUMPREC:	break
		fc = conn[(a,b)]
		fd = conn[(b,a)]
		_,_,c = arrangeface(faces[fc], a)
		_,_,d = arrangeface(faces[fd], b)
		faces[fc] = (d,a,c)
		faces[fd] = (c,b,d)
		registerface(conn, fc)
		registerface(conn, fd)
		# update scores
		update_scores((d,a))
		update_scures((a,c))
		update_scores((c,d))
		update_scores((d,a))
		update_scures((a,c))
		update_scores((c,d))
	
from .mesh import suites
from .nprint import nprint

from .mathutils import atan, mix
from . import settings
	
def discretise_refine(curve: '[(x, f(x))]', func: 'f(float) -> float', resolution=None, simplify=True):
	''' improve discretisation to reach a certain resolution '''
	if resolution and resolution[0] == 'div':
		raise ValueError("resolution must have a local meaning, so not 'div'")
	
	r = 1/2		# refinement placement, 1/3 ensure that the points are placed regularly
	pts = [curve[0], curve[1]]	# use a new list to improve insertion efficiencies (we will always insert close to the end)
	i = 2	# index in pts
	j = 2	# index in curve
	target = len(curve)
	
	# check curve around b
	def goodcorner(a, b, c):
		length = distance(a, b) + distance(b, c)
		d1 = (c[1] - b[1]) / (c[0] - b[0])
		d2 = (b[1] - a[1]) / (b[0] - a[0])
		angle = abs(atan(d1) - atan(d2))
		return settings.curve_resolution(length, angle, resolution) <= 1
	
	while True:
		# load following if necessary
		if i >= len(pts):
			if j >= target:	
				break
			pts.append(curve[j])
			j += 1
		
		# check resolution around i-1
		if not goodcorner(pts[i-2], pts[i-1], pts[i]):
			# refine around i-1
			r1 = mix(pts[i-1][0], pts[i][0],	r)
			r2 = mix(pts[i-1][0], pts[i-2][0],	r)
			p1 = (r1, func(r1))
			p2 = (r2, func(r2))
			if goodcorner(pts[i-2], pts[i-1], p1):
				pts.insert(i, p1)
			elif goodcorner(p2, pts[i-1], pts[i]):
				pts.insert(i-1, p2)
			else:
				pts.insert(i, 	p1)
				pts.insert(i-1,	p2)
		else:
			# go to next point
			i += 1
			
	# simplify overresolution
	i = 4
	while i < len(pts):
		if goodcorner(pts[i-4], pts[i-3], pts[i-1]) and goodcorner(pts[i-3], pts[i-1], pts[i]):
			pts.pop(i-2)
		else:
			i += 1
	
	return pts

	
def loop_closer(lines: Web, ending=0) -> 'Web':
	''' generate closing lines to create a loop.
	
		:ending:	the thickness along normals, for exterior geometries
	'''
	# the lines are simple lines, no junction at all
	if not line.isline():
		raise ValueError("the given web is not only made of lines, there is junctions")
	
	pts = lines.pts
	# create independant unclosed loops with parts cutted by bridges
	bridges = line_bridges(lines, normal)
	junctions = set()
	for a,b in bridges:
		junctions.add(a)
		junctions.add(b)
	parts = suites(parts)
	loops = []
	while parts:
		assembly = [parts.pop()]	# suites of points to extend
		for i,p in enumerate(assembly):
			# junction created by a bridge
			if p in bridges:
				for j,bridge in enumerate(bridges):
					if p in bridge:
						n = bridge[int(p == bridge[0])]
						bridges.pop(j)
						break
				else:
					continue
				n = bridges[p]
				# restack the rest of the current part
				parts.append(assembly[i+1:])
				assembly = assembly[:i+1]
				# search the junction index in the new part
				for part in parts:
					if n in part:	break
				else:
					raise Exception('algorithm failure: bridge with no matching river')
				j = part.index(n)
				# append the new part
				assembly += part[j:]
				if part[-1] == part[0]:	
					assembly += part[:j]
		loops.append(assembly)
	
	for loop in loops:
		# get a convex outline for the opened part, take not of which part is a real edge and what is bridge
		outline = convexhull(Wire(pts, reversed(loop)), start=loop[0], stop=loop[-1])
		# thicken the convex outline based on the curviline absciss
		pts.extend(pts[p] + thickness*normals[p] 	for p in outline[1:-1])
		l = len(pts)
		lines.append((outline[0], l))
		lines.append((l+len(outline), outline[-1]))
		lines.extend((i,i+1)	for i in range(len(outline)-1))



# This file is part of pymadcad,  distributed under license LGPL v3

'''	
	This module defines meshes composed of triangles or of edges.
	
	Architecture
	------------
	
	The data structures defined here are just containers, they do not intend a specific usage of the geometries as it can be with halfedges, binary trees, etc. The data structures here intend to store efficiently the mesh information and allow basic data operation like concatenation, find-replace, and so.
	
	The following considerations are common to the classes here:
	
	- the classes are just wrapping references to data buffers
	
		they do not own exclusively their content, nor it is forbidden to hack into their content (it is just lists, you can append, insert, copy, etc what you want).
		As so, it is almost no cost to shallow-copy a mesh, or to create it from existing buffers without computation nor data copy. It is very common that several meshes are sharing the same buffer of points, faces, groups ...
		
	- the management of the container's data ownership is left to the user
	
		the user has to copy the data explicitely when necessary to unlink containers sharing the same buffers. To allow the user to do so, the madcad functions observe the following rules:
		
		+ the container's methods that modify a small portion of its data does so in-place. when they do not return a particular value, they do return 'self' (lowercase, meaning the current container instance)
		
		+ the container's methods that modify a big amount of its data do return a new container instance, sharing the untouched buffers with the current one and duplicating the altered data. They do return a new instance of 'Self' (uppercase, meaning it is the current container type)
		
	- the methods defined for containers are only simple and very general operations. 
	
		The more complex operations are left to separated functions.
		Most of the time, container methods are for data/buffers management, connectivity operations, and mathematical caracteristics extractions (like normals, surface, volume, etc)
		
	See `Mesh.own()` for instance.
		
		
	Classes defined here
	--------------------
	
		:Mesh:	mesh of triangles
		:Web:	mesh of edges
		:Wire:	mesh of points, often used as a curve/polygon defined by a suite of points
		
	Topology
	--------
		
	A lot of similarities exists between these classes, because many of their methods are topologically generic to every dimension. They are often implemented for specific mesh
	For convenience, there is aliases to these classes highlighting their topology genericity
	
		:Mesh0:	(alias to `Wire`) mesh with inner dimension 0 (simplices are points)
		:Mesh1:	(alias to `Web`) mesh with inner dimension 1 (simplices are edges)
		:Mesh2:	(alias to `Mesh`) mesh with inner dimension 2 (simplices are triangles)
		
	Inner identification
	--------------------
	
	The containers are provided with an inner-identification system allowing to keep track of portion of geometries in the mesh across operations applies on. This is permitted by their field `tracks` and `groups`. Each simplex (point, edge or face depending of the mesh inner dimension) in meshes is associated to an identification number referencing a group definition in the group list. So faces can be filtered at any moment depedning on their group attributes or their group number.
	
	Groups definitions can be anything, but more often is a dictionnary (containing the group attributes we can filter on), or simply `None`
	
	See `Mesh.group()` for more details.
'''

from ..mathutils import *


__all__ = [
		'Mesh', 'Web', 'Wire', 'MeshError', 'web', 'wire', 
		'connpp', 'connpp', 'connpe', 'connef',
		'edgekey', 'facekeyo', 'arrangeface', 'arrangeedge', 
		'suites', 'line_simplification', 'mesh_distance', 'striplist',
		'typedlist_to_numpy', 'numpy_to_typedlist', 'ensure_typedlist',
		]


# submodules content exposed
from .mesh import Mesh, mkquad, mktri
from .web import Web
from .wire import Wire
from .conversions import *
from .container import (
	MeshError, NMesh,
	connpe, connef, connpp, connexity,
	facekeyo, edgekey, arrangeface, arrangeedge,
	suites, striplist,
	typedlist_to_numpy, numpy_to_typedlist, ensure_typedlist,
	)

# topological genericity definitions
Mesh0 = Wire
Mesh1 = Web
Mesh2 = Mesh


def line_simplification(web, prec=None):
	''' return a dictionnary of merges to simplify edges when there is points aligned.
	
		This function sort the points to remove on the height of the triangle with adjacent points.
		The returned dictionnary is guaranteed without cycles
	'''
	if not prec:	prec = web.precision()
	pts = web.points
	
	# simplify points when it forms triangles with too small height
	merges = {}
	def process(a,b,c):
		# merge with current merge point if height is not sufficient, use it as new merge point
		height = length(noproject(pts[c]-pts[b], pts[c]-pts[a]))
		if height > prec:
			#scn3D.add(text.Text(pts[b], str(height), 8, (1,0,1), align=('left', 'center')))
			return b
		else:
			merges[b] = a
			return a
	
	for k,line in enumerate(suites(web.edges, oriented=False)):
		s = line[0]
		for i in range(2, len(line)):
			s = process(s, line[i-1], line[i])
		if line[0]==line[-1]: process(s, line[0], line[1])
		
	# remove redundancies in merges (there can't be loops in merges)
	for k,v in merges.items():
		while v in merges and merges[v] != v:
			merges[k] = v = merges[v]
	return merges


def distance2_pm(point, mesh) -> '(d, prim)':
	''' squared distance from a point to a mesh
	'''
	score = inf
	best = None
	if isinstance(mesh, Mesh):
		for face in mesh.faces:
			f = mesh.facepoints(face)
			n = cross(f[1]-f[0], f[2]-f[0])
			if not n:	continue
			# check if closer to the triangle's edges than to the triangle plane
			plane = True
			for i in range(3):
				d = f[i-1]-f[i-2]
				if dot(cross(n, d), point-f[i-2]) < 0:
					x = dot(point-f[i-2], d) / length2(d)
					# check if closer to the edge points than to the edge axis
					if x < 0:	dist, candidate = distance2(point, f[i-2]), face[i-2]
					elif x > 1:	dist, candidate = distance2(point, f[i-1]), face[i-1]
					else:		dist, candidate = length2(noproject(point - f[i-2], d)), (face[i-2], face[i-1])
					plane = False
					break
			if plane:
				dist, candidate = dot(point-f[0], n) **2 / length2(n), face
			if dist < score:
				best, score = candidate, dist
	elif isinstance(mesh, (Web,Wire)):
		if isinstance(mesh, Web):	edges = mesh.edges
		else:						edges = mesh.edges()
		for edge in edges:
			e = mesh.edgepoints(edge)
			d = e[1]-e[0]
			x = dot(point - e[0], d) / length2(d)
			# check if closer to the edge points than to the edge axis
			if x < 0:	dist, candidate = distance2(point, e[0]), edge[0]
			elif x > 1:	dist, candidate = distance2(point, e[1]), edge[1]
			else:		dist, candidate = length2(noproject(point - e[0], d)), edge
			if dist < score:
				best, score = candidate, dist
	elif isinstance(mesh, vec3):
		return distance2(point, mesh), 0
	else:
		raise TypeError('cannot evaluate distance from vec3 to {}'.format(type(mesh)))
	return score, best

def mesh_distance(m0, m1) -> '(d, prim0, prim1)':
	''' minimal distance between elements of meshes 
	
		The result is a tuple `(distance, primitive from m0, primitive from m1)`.
		`primitive` can be:
		
			:int:				index of the closest point
			:(int,int):			indices of the closest edge
			:(int,int,int): 	indices of the closest triangle
	'''
	# compute distance from each points of m to o
	def analyse(m, o):
		# get an iterator over actually used points only
		if isinstance(m, Mesh):
			usage = [False]*len(m.points)
			for f in m.faces:
				for p in f:	usage[p] = True
			it = (i for i,u in enumerate(usage) if u)
		elif isinstance(m, Web):
			usage = [False]*len(m.points)
			for e in m.edges:
				for p in e:	usage[p] = True
			it = (i for i,u in enumerate(usage) if u)
		elif isinstance(m, Wire):
			it = m.indices
		elif isinstance(m, vec3):
			return (*distance2_pm(m, o), 0)
		# comfront to the mesh
		return min((
				(*distance2_pm(m.points[i], o), i)
				for i in it), 
				key=lambda t:t[0])
	# symetrical evaluation
	d0 = analyse(m0, m1)
	d1 = analyse(m1, m0)
	if d0[0] < d1[0]:	return (sqrt(d0[0]), d0[2], d0[1])
	else:				return (sqrt(d1[0]), d1[1], d1[2])
	



from math import inf
from .mathutils import glm,vec2,dvec2,dmat2, perp, perpdot, length, distance, normalize, inverse, isnan, norminf, NUMPREC
from .mesh import Mesh, Web, Wire


def triangulation(outline, prec=NUMPREC):
	try:				m = triangulation_outline(outline, prec)
	except MeshError:	m = triangulation_skeleton(outline, prec)
	return m


def skeleting(outline: Wire, skeleting: callable, prec=NUMPREC) -> [vec2]:
	''' skeleting procedure for the given wire
		at each step, the skeleting function is called
		created points will be added to the wire point buffer and this buffer is returned (ROI)
	'''
	l = len(outline)
	pts = outline.points
	
	# edge normals
	enormals = [perp(normalize(outline[i]-outline[i-1]))	for i in range(l)]
	# compute half axis starting from each point
	haxis = []
	for i in range(l):
		haxis.append((outline.indices[i], i, (i+1)%l))
	
	# create the intersections to update
	intersect = [(0,0)] * l	# intersection for each edge
	dist = [-1] * l	# min distance for each edge
	def eval_intersect(i):
		o1,a1,b1 = haxis[i-1]
		o2,a2,b2 = haxis[i]
		# compute normals to points
		v1 = enormals[a1]+enormals[b1]
		v2 = enormals[a2]+enormals[b2]
		if norminf(v1) < prec:	v1 =  perp(enormals[b1])	# if edges are parallel, take the direction to the shape
		if norminf(v2) < prec:	v2 = -perp(enormals[b2])
		# compute the intersection
		x1,x2 = inverse(dmat2(v1,-v2)) * dvec2(-pts[o1] + pts[o2])
		if x1 >= -NUMPREC and x2 >= -NUMPREC:
			intersect[i] = pts[o2] + x2*v2
			dist[i] = min(x1*length(v1), x2*length(v2))
		elif isnan(x1) or isnan(x2):
			intersect[i] = pts[o2]
			dist[i] = 0
		else:
			intersect[i] = None
			dist[i] = inf
	for i in range(l):
		eval_intersect(i)
	
	# build skeleton
	while len(haxis) > 1:
		print(dist)
		i = min(range(len(haxis)), key=lambda i:dist[i])
		assert dist[i] != inf, "no more intersection found (algorithm failure)"
		o1,a1,b1 = haxis[i-1]
		o2,a2,b2 = haxis[i]
		# add the intersection point
		ip = len(pts)
		pts.append(intersect[i])
		# extend skeleton
		skeleting(haxis, i, ip)
		# create the new half axis
		haxis.pop(i)
		dist.pop(i)
		intersect.pop(i)
		haxis[i-1] = (ip, a1, b2)
		eval_intersect((i-2) % len(haxis))
		eval_intersect((i-1) % len(haxis))
		eval_intersect(i % len(haxis))
		eval_intersect((i+1) % len(haxis))
		eval_intersect((i+2) % len(haxis))
	
	return pts

def skeleton(outline: Wire, prec=NUMPREC) -> Web:
	''' return a Web that constitute the skeleton of the outline
		the returned Web uses the same point buffer than the input Wire.
		created points will be added into it
	'''
	skeleton = []
	#def sk(ip, o1,o2, a1,b1, a2,b2):
		#skeleton.append((o1,ip))
		#skeleton.append((o2,ip))
	def sk(haxis, i, ip):
		skeleton.append((haxis[i-1][0], ip))
		skeleton.append((haxis[i][0], ip))
	pts = skeleting(outline, sk, prec)
	return Web(pts, skeleton)

def triangulation_skeleton(outline: Wire, prec=NUMPREC) -> Mesh:
	''' return a Mesh with triangles filling the surface of the outline 
		the returned Mesh uses the same point buffer than the input Wire
	'''
	triangles = []
	skeleton = []
	pts = outline.points
	original = len(pts)
	minbone = [inf]
	def sk(haxis, i, ip):
		triangles.append((haxis[i-2][0], haxis[i-1][0], ip))
		triangles.append((haxis[i-1][0], haxis[i][0], ip))
		triangles.append((haxis[i][0], haxis[(i+1)%len(haxis)][0], ip))
		if   haxis[i-1][0] < original:
			d = distance(pts[haxis[i-1][0]], pts[ip])
			if d < minbone[0]:	minbone[0] = d
		else:
			skeleton.append((haxis[i-1][0], ip))
		if haxis[i][0] < original:
			d = distance(pts[haxis[i][0]], pts[ip])
			if d < minbone[0]:	minbone[0] = d
		else:
			skeleton.append((haxis[i][0], ip))
	pts = skeleting(outline, sk, prec)
	m = Mesh(pts, triangles)
	# merge points from short internal edges
	minbone = 0.5*minbone[0]
	merges = {}
	for a,b in skeleton:
		if distance(pts[a], pts[b]) < minbone:
			if   a not in merges:	merges[a] = merges.get(b,b)
			elif b not in merges:	merges[b] = merges.get(a,a)
	for k,v in merges.items():
		while v in merges and merges[v] != v:	
			if distance(pts[k], pts[v]) > minbone:
				merges[k] = k
				merges[v] = v
				break
			merges[k] = v = merges[v]
	m.mergepoints(merges)
	return m

def triangulation_outline(outline: Wire) -> Mesh:
	''' return a mesh with the triangles formed in the outline
		the returned mesh uses the same buffer of points than the input
	'''
	pts = outline.points
	
	hole = list(outline.indices)
	scores = [score(i) for i in range(len(outline))]
	def score(i):
		u = pts[hole[i+1]] - pts[hole[i]]
		v = pts[hole[i-1]] - pts[hole[i]]
		score = perpdot(u,v) / (length(u)+length(v)+length(u-v))
		if score < 0:	return score
		# check that there is not point of the outline inside the triangle
		decomp = inverse(dmat2(u,v))
		inside = False
		o = pts[hole[i]]
		for j,ip in range(len(hole)):
			if j != i and j != (i-1)%l and j != (i+1)%l:
				p = pts[hole[j]]
				uc,vc = decomp * dvec2(p-o)
				if 0 < uc and 0 < vc and uc+vc < 1:
					inside = True
					break
		if inside:
			score = -inf
		return score
	
	triangles = []
	while len(hole) > 2:
		l = len(outline)
		i = max(range(l), key=lambda i:scores[i])
		assert scores[i] > 0, "no more feasible triangles (algorithm failure)"
		triangles.append((hole[(i-1)%l], hole[i], hole[(i+1)%l]))
		hole.pop(i)
		scores[i-1] = score(i-1)
		scores[i] = score(i)
	
	return Mesh(pts, triangles)

from .mathutils import dot

def triangulation_sweepline(outline: Web) -> Mesh:
	''' sweep line algorithm to triangulate an outline
		the web edges should not be oriented, and thus the resulting face has no orientation
		complexity: O(n*ln2(n))
	'''
	pts = outline.points
	faces = []
	# select the sweepline direction with the maximum interval
	diag = glm.abs(outline.box().width)
	sortdim = 0
	if diag[1] > sortdim:	sortdim = 1
	if diag[2] > sortdim:	sortdim = 2
	# direction of direct rotation of the outline, orthogonal to the sweepline direction
	y = normalize(pts[0]-pts[1])
	y[sortdim] = 0
	# orient edges along the axis and sort them
	if len(outline.edges) < 3:	return Mesh()
	edges = outline.edges[:]
	for i,(a,b) in enumerate(edges):
		if pts[a][sortdim] < pts[b][sortdim]:	edges[i] = b,a
	stack = sorted(edges, key=lambda e: (pts[e[0]][sortdim], dot(pts[e[1]]-pts[e[0]], y)/(pts[e[1]][sortdim]-pts[e[0]][sortdim] or 1e-9) ) )
	# direction of direct rotation of the outline, orthogonal to the sweepline direction
	#y = pts[stack[-1][1]] - pts[stack[-2][1]]
	#y[sortdim] = 0
	
	def affine(e):
		v = pts[e[1]] - pts[e[0]]
		return (pts[e[0]], v/v[sortdim] if v[sortdim] else 0)
		
	clusters = []
	while stack:
		edge = stack.pop()
		# get the pair edge if there is one starting at the same point
		coedge = None
		for i in reversed(range(len(stack))):
			if pts[stack[i][0]][sortdim] < pts[edge[0]][sortdim]:	break
			#if edge[0] == stack[i][0]:
			if edge[0] == stack[i][0] and edge[1] != stack[i][1]:
				coedge = stack.pop(i)
				if dot(pts[edge[1]]-pts[edge[0]], y) < dot(pts[coedge[1]]-pts[coedge[0]], y):
					edge, coedge = coedge, edge
				break
		print(edge, coedge)
		
		# search in which cluster we are
		found = False
		#for i,cluster in enumerate(clusters):
		i = 0
		while i < len(clusters):
			p = pts[edge[0]]
			l0,l1,a0,a1 = clusters[i]
			
			if l0[1] == l1[1]:	
				print('    pop', l0,l1)
				faces.append((l0[0], l0[1], l1[0]))
				clusters.pop(i)
				continue
			
			print('   ', l0,l1,edge[0])
			# interior hole
			if coedge and dot(p - a0[0]+a0[1]*p[sortdim], y) < 0 and dot(p - a1[0]+a1[1]*p[sortdim], y) > 0:
					print('      hole for ',i, edge)
					faces.append((l1[0], l0[0], edge[0]))
					clusters[i] = (l0, edge, a0, affine(edge))
					clusters.insert(i, (coedge, l1, affine(coedge), a1))
					found = True
					break
			# continuation of already existing edge of the cluster
			elif edge[0] == l0[1]:
				if coedge:
					clusters.insert(i+1, (edge, coedge, affine(edge), affine(coedge)))
					print('      ramif')
				else:
					coedge = edge
					print('      continuation')
				faces.append((l1[0], l0[0], l0[1]))
				clusters[i] = (coedge, l1, affine(coedge), a1)
				found = True
				break
			elif edge[0] == l1[1]:
				faces.append((l0[0], l1[1], l1[0]))
				clusters[i] = (l0, edge, a0, affine(edge))
				if coedge:
					print('      ramif')
					clusters.insert(i, (edge, coedge, affine(edge), affine(coedge)))
				else:
					print('      continuation')
				found = True
				break
			i += 1
		# if it's a new corner
		if coedge and not found:
			print('    new cluster')
			c = (edge, coedge, affine(edge), affine(coedge))
			j = 0
			for j in range(len(clusters)):
				ci, cp = c[2], clusters[j][3]
				if (ci[0][sortdim], dot(ci[1],y)/(ci[1][sortdim] or 1e-9)) >= (cp[0][sortdim], dot(cp[1],y)/(cp[1][sortdim] or 1e-9)):
					break
			faces.append((edge[0], edge[1], coedge[1]))
			clusters.insert(j, c)
			from nprint import nprint
			nprint(clusters)
		#elif not coedge and not found:
			#raise Exception("algorithm failure, can be due to the given outline")
			
	for cluster in clusters:
		l0,l1,a0,a1 = cluster
		faces.append((l0[0], l0[1], l1[0]))
		if l0[1] != l1[1]:
			faces.append((l1[0], l0[1], l1[1]))
	
	return Mesh(pts, faces)



import numpy.core as np
from .mathutils import *
from .rendering import Display

class MeshEditor:
	def __init__(self, scene, mesh=None):
		self.points = np.empty((0,3), dtype='f4')
		self.faces = np.empty((0,3), dtype='u4')
		self.pointflags = np.empty((0,), dtype='u1')
		self.faceflags = np.empty((0,), dtype='u1')
		self.vb_points = ctx.buffer(self.points)
		self.vb_faces = ctx.buffer(self.faces)
		self.vb_edges = ctx.buffer(self.edges)
		self.vb_pointflags = ctx.buffer(self.pointflags)
		self.vb_faceflags = ctx.buffer(self.faceflags)
		
		self.va_points = ctx.vertex_array(
			self.pointshader,
			[	(self.vb_points, '3f4', 'v_position'),
				(self.vb_pointflags, 'u1', 'v_flag')],
			)
		self.va_faces = ctx.vertex_array(
			self.faceshaders,
			[(self.vb_points, '3f4', 'v_position')],
			)
		indev



# This file is part of pymadcad,  distributed under license LGPL v3

import math
from .mathutils import (
		vec3, uvec2,
		anglebt, cross, dot, length, distance, normalize, noproject,
		distance_pp, distance_pa, distance_pe, distance_aa, distance_ae,
		)
from .mesh import Mesh, Web, edgekey, connpp, connef

__all__ = ['select', 'stopangle', 'crossover', 'straight', 'short', 'selexpr', 'edgenear']

def select(mesh, edge, stopleft=None, stopright=False, conn=None, web=None) -> Web:
	''' Select edges by propagation across group outlines, until the stop criterion is verified
		If a criterion is None, the propagation select everything it reachs.
		If stopright is not specified (False), stopleft is used for both directions.
		
		example:
			select(m, (12,37), stopangle(pi/2) | crossover)
	'''
	# manage arguments
	if isinstance(mesh, Mesh) and web is None:
		web = mesh.groupoutlines()
	elif isinstance(mesh, Web):
		web = mesh
		mesh = None
	if conn is None:
		conn = connpp(web.edges)
	if stopleft is None:		stopleft = lambda *args: False
	if stopright is None:		stopright = lambda *args: False
	elif stopright is False:	stopright = stopleft
	
	if isinstance(edge, vec3):
		edge = edgenear(web, edge)
	
	assert edge[0] in conn and edge[1] in conn[edge[0]], "the given edge doesn't exist"
	# selection by propagation on each side
	shared = {'mesh':mesh, 'web':web, 'conn':conn}
	seen = set()
	selectside(shared, edge, stopleft, seen, True) 
	selectside(shared, edge, stopleft, seen, False) 
	return Web(web.points, list(seen))

def selectside(shared, edge, stop, seen=None, revert=None):
	''' selection by propagation until stop criterion '''
	front = [edge if not revert else uvec2(edge[1],edge[0])]
	if seen is None:	
		seen = set()
	else:				
		seen.discard(edge)
		seen.discard(uvec2(edge[1], edge[0]))
	conn = shared['conn']
	while front:
		last,curr = front.pop()
		if   uvec2(last,curr) in seen:		continue
		elif uvec2(curr,last) in seen:		continue
		if revert:	seen.add(uvec2(curr,last))
		else:		seen.add(uvec2(last,curr))
		connected = conn[curr]
		if len(connected) > 1:
			for next in connected:
				if not stop(shared,last,curr,next):
					front.append(uvec2(curr,next))
		else:
			front.append(uvec2(curr,connected[0]))
	return seen


class selexpr(object):
	''' boolean expression for topology condition, used to define stop expressions for select() '''
	__slots__ = 'func',
	
	def __init__(self, func):
		self.func = func
	
	def __and__(self, other):	return selexpr(lambda *args: self.func(*args) and other.func(*args))
	def __or__(self, other):	return selexpr(lambda *args: self.func(*args) or other.func(*args))
	def __xor__(self, other):	return selexpr(lambda *args: bool(self.func(*args) ^ other.func(*args)))
	def __not__(self, other):	return selexpr(lambda *args: not self.func(*args))
	def __call__(self, *args):	return self.func(*args)
	
	def __repr__(self):
		return 'selexpr({})'.format(self.func)

# --- stop conditions ---

def stopangle(maxangle):
	''' stop when angle between consecutive edges is bigger than maxangle '''
	def stop(shared,last,curr,next): 
		points = shared['web'].points
		return anglebt(points[curr]-points[last], points[next]-points[curr]) >= maxangle
	return selexpr(stop)

@selexpr
def crossover(shared,last,curr,next):
	return len(shared['conn'][curr]) > 2

@selexpr
def straight(shared,last,curr,next):
	points = shared['web'].points
	start = points[curr] - points[last]
	other = min(shared['conn'][curr], key=lambda o: anglebt(start, points[o] - points[curr]))
	return other != next

@selexpr
def short(shared,last,curr,next):
	points = shared['web'].points
	start = points[curr] - points[last]
	best = None
	score = 0
	for o in shared['conn'][curr]:
		if o == last:	continue
		angle = anglebt(start, points[o] - points[curr])
		if angle > score:
			score = angle
			best = o
	return best != next

def faceangle(minangle):
	''' stop when angle between adjacent faces is strictly lower than minangle '''
	def stop(shared,last,curr,next):
		mesh = shared['mesh']
		if 'connef' not in shared:
			shared['connef'] = connef = connef(context['mesh'])
		else:
			connef = shared['connef']
		if (curr,next) in connef and (next,curr) in connef:
			nl = mesh.facenormal(connef[(curr,next)])
			nr = mesh.facenormal(connef[(next,curr)])
			return anglebt(nl, nr) <= minangle
		else:
			return True
	return selexpr(stop)
	
# --- edge selection from mesh ---

def edgenear(web, obj):
	if isinstance(obj,vec3):	dist = lambda e: distance_pe(obj,e)
	elif isinstance(obj,tuple):	dist = lambda e: distance_ae(obj,e)
	else:
		raise TypeError("obj must be a point or an axis")
	if isinstance(web, Mesh):	web = web.groupoutlines()
	best = None
	score = math.inf
	for edge in web.edges:
		d = dist((web.points[edge[0]], web.points[edge[1]]))
		if d < score:
			score = d
			best = edge
	if isinstance(obj,tuple) and dot(web.points[best[1]]-web.points[best[0]], obj[1]) < 0:
		best = best[1],best[0]
	return best



# This file is part of pymadcad,  distributed under license LGPL v3
'''	
	This file is to write text into the 3D view
	
	The current implementation uses a bitmap font texture baked at program start.
	It's following the first technique described here:
		https://learnopengl.com/In-Practice/Text-Rendering
'''

from PIL import Image, ImageFont, ImageDraw
import numpy.core as np
import moderngl as mgl
from .common import ressourcedir
from .mathutils import fvec3, fvec4, fmat4, ceil, sqrt
from . import settings, rendering

# TODO: utiliser les methodes et attributs ImgeFont.size, .getmetrics(), etc pour avoir les hauteur et largeur de police

def create_font_texture(font, maxchar=1100):
	# determine the size of the needed texture
	fontsize = font.size + 4
	width = 64
	while 2*(width//fontsize)**2 < maxchar:
		width *= 2
	c = width//(fontsize//2)
	l = maxchar//c +1
	width = c * (fontsize//2)
	height = l * fontsize
	# create texture
	tex = Image.new('L', (width,height), 0)
	draw = ImageDraw.Draw(tex)
	# draw the font onto
	for i in range(maxchar):
		draw.text(char_placement(font.size, c, l, i), chr(i), fill=255, font=font)
	#tex.show()
	
	return tex, (c, l)

class Text:
	def __init__(self, position, text, size=None, color=(1,1,1), align=(0,0)):
		self.text = text
		self.position = fvec3(position)
		self.size = size or settings.display['view_font_size']
		self.color = color
		self.align = align
	
	def display(self, scene):
		return TextDisplay(scene, self.position, self.text, self.size, self.color, self.align)

def char_placement(fontsize, c, l, n):
	fontsize += 4
	return ((fontsize//2) * (n%c), fontsize * (n//c))

pointsdef = [
	[0, 0],
	[0,-1],
	[1,-1],
	[0, 0],
	[1,-1],
	[1, 0],
	]

class TextDisplay(rendering.Display):
	
	def __init__(self, scene, position, text, size=None, color=None, align=(0,0), layer=0):
		if not color:	color = settings.display['annotation_color']
		if not size:	size = settings.display['view_font_size']
		self.position = fvec3(position)
		self.color = fvec3(color)
		self.size = size
		self.layer = layer
		
		# load font
		def load(scene):
			img, align = create_font_texture(ImageFont.truetype(ressourcedir+'/NotoMono-Regular.ttf', 2*size))
			return scene.ctx.texture(img.size, 1, img.tobytes()), align
		self.fonttex, self.fontalign = scene.ressource(('fonttex', size), load)
		
		# load shader
		def load(scene):
			shader = scene.ctx.program(
						vertex_shader=open(ressourcedir+'/shaders/font.vert').read(),
						fragment_shader=open(ressourcedir+'/shaders/font.frag').read(),
						)
			shader['fonttex'].value = 0
			return shader
		self.shader = scene.ressource('shader_font', load)
		
		# place triangles
		points = np.zeros((len(pointsdef)*len(text), 4), 'f4')
		l = 0
		c = 0
		for i,char in enumerate(text):
			if char == '\n':
				c = 0
				l += 1
			elif char == '\t':
				c += 4 - c%4	# TODO: use a tab size in settings
			elif char == ' ':
				c += 1
			else:
				n = ord(char)
				#placement = char_placement(2*size, *self.fontalign, n)
				for j,add in enumerate(pointsdef):
					points[len(pointsdef)*i + j] = [
						add[0]+c, 
						add[1]-l, 
						(n % self.fontalign[0] +add[0]) / self.fontalign[0], 
						(n //self.fontalign[0] -add[1]) / self.fontalign[1],
						]
				c += 1
		l += 1		
		align = (processalign(align[0], c), -processalign(align[1], l))
		points -= [*align, 0, 0]
		self.textsize = (c,l)
		self.visualsize = (-align[0], -align[1], c//2-align[0], l-align[1])
			
		
		# create the buffers on the GPU
		self.vb_points = scene.ctx.buffer(points)
		self.vb_faces = scene.ctx.buffer(np.arange(points.shape[0], dtype='u4'))
		self.va = scene.ctx.vertex_array(
			self.shader,
			[(self.vb_points, '2f 2f', 'v_position', 'v_uv')],
			)
	
	def render(self, view):		
		self.shader['layer'] = self.layer
		self.shader['color'].write(self.color)
		self.shader['position'].write(fvec3(self.world * fvec4(self.position,1)))
		self.shader['view'].write(view.uniforms['view'])
		self.shader['proj'].write(view.uniforms['proj'])
		self.shader['ratio'].value = (
				(self.size-2.5) / view.width()*2,
				(self.size-2.5) / view.height()*4,
				)
		self.fonttex.use(0)
		self.va.render(mgl.TRIANGLES)

	def stack(self, view):
		return ((), 'screen', 2, self.render),

def processalign(align, size):
	if isinstance(align, str):
		if align == 'left':		return 0
		elif align == 'right':	return size
		elif align == 'center':	return size/2
		else:
			raise ValueError("align must be int, float or any of 'left', 'right', 'center'")
	else:
		return align

def textsize(text, tab=4):
	''' visual size of a text as vec2(column,line) '''
	l = 0
	c = 0
	for i,char in enumerate(text):
		if char == '\n':
			c = 0
			l += 1
		elif char == '\t':
			c += tab - c%tab
		else:
			c += 1
	l += 1
	return vec2(c,l)

#class Particles(rendering.Display):
	#def __init__(self, texture, format, vertices):
		#position, tile, size



'''
	Soft(
		points=[(vec3, strength, smooth)],		smooth: face, line, fixed
		lines=[(indices)],
		faces=[((indices), (indices))],
		spec={p:normal, e:tangent},
		)
'''

from .mathutils import *
from .mesh import MeshError

class Soft:
	def __init__(self, points=None, attrs=None, lines=None, faces=None, spec=None, groups=None):
		self.points = points or []
		self.attrs = attrs or [(1, 0)] * len(self.points)
		self.lines = lines or []
		self.faces = faces or []
		self.spec = spec or {}
		self.groups = groups or [None] * len(self.faces)
		
	def check(self):
		l = len(self.points)
		for line in self.lines:
			for i in line:
				if i >= l:		raise MeshError("some line indices are greater than the number of points")
				if i < 0:		raise MeshError("point indices must be positive")
		if len(self.groups) != len(self.faces):	raise MeshError("the number of group and faces are different")
		if len(self.points) != len(self.attrs):	raise MeshError("the number of points and point attributes are different")
		
	def isvalid(self):
		try:	self.check()
		except MeshError:	return False
		return True
		
	def mesh(self, resolution=None):
		indev
		# choose discretization for each edge
		# make discretization coherent
		# mark specific lines
		# discretize face lines



def soft(mesh):
	''' convert a mesh to a Soft '''
	if isinstance(mesh, Soft):		return mesh
	elif isinstance(mesh, Mesh):	return Soft(points=mesh.points, faces=mesh.faces, groups=mesh.groups)
	elif isinstance(mesh, Web):		return Soft(points=mesh.points, lines=mesh.edges)
	else:
		raise TypeError('cannot convert {} to Soft'.format(type(mesh).__name))
